@@ -27,6 +27,7 @@
 
 #include "common/av_common.h"
 #include "common/common.h"
+#include "demux.h"
 
 #include "packet.h"
 
@@ -34,6 +35,7 @@ static void packet_destroy(void *ptr)
 {
     struct demux_packet *dp = ptr;
     av_packet_unref(dp->avpacket);
+    mp_packet_tags_unref(dp->metadata);
 }
 
 // This actually preserves only data and side data, not PTS/DTS/pos/etc.
@@ -54,6 +56,7 @@ struct demux_packet *new_demux_packet_from_avpacket(struct AVPacket *avpkt)
         .end = MP_NOPTS_VALUE,
         .stream = -1,
         .avpacket = talloc_zero(dp, AVPacket),
+        .kf_seek_pts = MP_NOPTS_VALUE,
     };
     av_init_packet(dp->avpacket);
     int r = -1;
@@ -74,6 +77,19 @@ struct demux_packet *new_demux_packet_from_avpacket(struct AVPacket *avpkt)
     return dp;
 }
 
+// (buf must include proper padding)
+struct demux_packet *new_demux_packet_from_buf(struct AVBufferRef *buf)
+{
+    if (!buf)
+        return NULL;
+    AVPacket pkt = {
+        .size = buf->size,
+        .data = buf->data,
+        .buf = buf,
+    };
+    return new_demux_packet_from_avpacket(&pkt);
+}
+
 // Input data doesn't need to be padded.
 struct demux_packet *new_demux_packet_from(void *data, size_t len)
 {
@@ -109,11 +125,13 @@ void demux_packet_copy_attribs(struct demux_packet *dst, struct demux_packet *sr
     dst->dts = src->dts;
     dst->duration = src->duration;
     dst->pos = src->pos;
+    dst->segmented = src->segmented;
     dst->start = src->start;
     dst->end = src->end;
-    dst->new_segment = src->new_segment;
+    dst->codec = src->codec;
     dst->keyframe = src->keyframe;
     dst->stream = src->stream;
+    mp_packet_tags_setref(&dst->metadata, src->metadata);
 }
 
 struct demux_packet *demux_copy_packet(struct demux_packet *dp)
