@@ -39,6 +39,10 @@
 
 #include "config.h"
 
+#if HAVE_MACOS_COCOA_CB
+#include "osdep/macOS_swift.h"
+#endif
+
 @interface EventsResponder ()
 {
     struct input_ctx *_inputContext;
@@ -194,11 +198,6 @@ void cocoa_put_key(int keycode)
     [[EventsResponder sharedInstance] putKey:keycode];
 }
 
-void cocoa_put_key_event(void *event)
-{
-    [[EventsResponder sharedInstance] handleKey:event];
-}
-
 void cocoa_put_key_with_modifiers(int keycode, int modifiers)
 {
     keycode |= [[EventsResponder sharedInstance] mapKeyModifiers:modifiers];
@@ -309,10 +308,13 @@ void cocoa_set_mpv_handle(struct mpv_handle *ctx)
 - (BOOL)setMpvHandle:(struct mpv_handle *)ctx
 {
     if (_is_application) {
-        dispatch_sync(dispatch_get_main_queue(), ^{ _ctx = ctx; });
+        dispatch_sync(dispatch_get_main_queue(), ^{
+            _ctx = ctx;
+            [NSApp setMpvHandle:ctx];
+        });
         return YES;
     } else {
-        mpv_detach_destroy(ctx);
+        mpv_destroy(ctx);
         return NO;
     }
 }
@@ -331,17 +333,21 @@ void cocoa_set_mpv_handle(struct mpv_handle *ctx)
 
 -(void)processEvent:(struct mpv_event *)event
 {
+    if(_is_application) {
+        [NSApp processEvent:event];
+    }
+
     switch (event->event_id) {
     case MPV_EVENT_SHUTDOWN: {
-        mpv_detach_destroy(_ctx);
+        #if HAVE_MACOS_COCOA_CB
+        if ([(Application *)NSApp cocoaCB].isShuttingDown)
+            return;
+        #endif
+        mpv_destroy(_ctx);
         _ctx = nil;
         break;
     }
     }
-
-    if(_is_application) {
-        [NSApp processEvent:event];
-    }
 }
 
 - (void)startAppleRemote
