@@ -1,20 +1,18 @@
 /*
  * This file is part of mpv.
  *
- * mpv is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * mpv is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * mpv is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License along
- * with mpv.  If not, see <http://www.gnu.org/licenses/>.
- *
- * Almost LGPL.
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with mpv.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <limits.h>
@@ -28,19 +26,19 @@
 #include <libavutil/rational.h>
 #include <libavcodec/avcodec.h>
 
+#if LIBAVUTIL_VERSION_MICRO >= 100
+#include <libavutil/mastering_display_metadata.h>
+#endif
+
 #include "mpv_talloc.h"
 
 #include "config.h"
+#include "common/av_common.h"
 #include "common/common.h"
+#include "hwdec.h"
 #include "mp_image.h"
 #include "sws_utils.h"
 #include "fmt-conversion.h"
-#include "gpu_memcpy.h"
-
-#include "video/filter/vf.h"
-
-#define HAVE_OPAQUE_REF (LIBAVUTIL_VERSION_MICRO >= 100 && \
-                         LIBAVUTIL_VERSION_INT >= AV_VERSION_INT(55, 47, 100))
 
 const struct m_opt_choice_alternatives mp_spherical_names[] = {
     {"auto",        MP_SPHERICAL_AUTO},
@@ -67,7 +65,7 @@ static int mp_image_layout(int imgfmt, int w, int h, int stride_align,
 
     // Note: for non-mod-2 4:2:0 YUV frames, we have to allocate an additional
     //       top/right border. This is needed for correct handling of such
-    //       images in filter and VO code (e.g. vo_vdpau or vo_opengl).
+    //       images in filter and VO code (e.g. vo_vdpau or vo_gpu).
 
     for (int n = 0; n < MP_MAX_PLANES; n++) {
         int alloc_w = mp_chroma_div_up(w, desc.xs[n]);
@@ -77,7 +75,7 @@ static int mp_image_layout(int imgfmt, int w, int h, int stride_align,
         out_plane_size[n] = out_stride[n] * alloc_h;
     }
     if (desc.flags & MP_IMGFLAG_PAL)
-        out_plane_size[1] = MP_PALETTE_SIZE;
+        out_plane_size[1] = AVPALETTE_SIZE;
 
     int sum = 0;
     for (int n = 0; n < MP_MAX_PLANES; n++) {
@@ -210,6 +208,10 @@ static void mp_image_destructor(void *ptr)
         av_buffer_unref(&mpi->bufs[p]);
     av_buffer_unref(&mpi->hwctx);
     av_buffer_unref(&mpi->icc_profile);
+    av_buffer_unref(&mpi->a53_cc);
+    for (int n = 0; n < mpi->num_ff_side_data; n++)
+        av_buffer_unref(&mpi->ff_side_data[n].buf);
+    talloc_free(mpi->ff_side_data);
 }
 
 int mp_chroma_div_up(int size, int shift)
@@ -298,6 +300,15 @@ void mp_image_unref_data(struct mp_image *img)
     }
 }
 
+static void ref_buffer(bool *ok, AVBufferRef **dst)
+{
+    if (*dst) {
+        *dst = av_buffer_ref(*dst);
+        if (!*dst)
+            *ok = false;
+    }
+}
+
 // Return a new reference to img. The returned reference is owned by the caller,
 // while img is left untouched.
 struct mp_image *mp_image_new_ref(struct mp_image *img)
@@ -312,27 +323,20 @@ struct mp_image *mp_image_new_ref(struct mp_image *img)
     talloc_set_destructor(new, mp_image_destructor);
     *new = *img;
 
-    bool fail = false;
-    for (int p = 0; p < MP_MAX_PLANES; p++) {
-        if (new->bufs[p]) {
-            new->bufs[p] = av_buffer_ref(new->bufs[p]);
-            if (!new->bufs[p])
-                fail = true;
-        }
-    }
-    if (new->hwctx) {
-        new->hwctx = av_buffer_ref(new->hwctx);
-        if (!new->hwctx)
-            fail = true;
-    }
+    bool ok = true;
+    for (int p = 0; p < MP_MAX_PLANES; p++)
+        ref_buffer(&ok, &new->bufs[p]);
 
-    if (new->icc_profile) {
-        new->icc_profile = av_buffer_ref(new->icc_profile);
-        if (!new->icc_profile)
-            fail = true;
-    }
+    ref_buffer(&ok, &new->hwctx);
+    ref_buffer(&ok, &new->icc_profile);
+    ref_buffer(&ok, &new->a53_cc);
 
-    if (!fail)
+    new->ff_side_data = talloc_memdup(NULL, new->ff_side_data,
+                        new->num_ff_side_data * sizeof(new->ff_side_data[0]));
+    for (int n = 0; n < new->num_ff_side_data; n++)
+        ref_buffer(&ok, &new->ff_side_data[n].buf);
+
+    if (ok)
         return new;
 
     // Do this after _all_ bufs were changed; we don't want it to free bufs
@@ -365,6 +369,10 @@ struct mp_image *mp_image_new_dummy_ref(struct mp_image *img)
     for (int p = 0; p < MP_MAX_PLANES; p++)
         new->bufs[p] = NULL;
     new->hwctx = NULL;
+    new->icc_profile = NULL;
+    new->a53_cc = NULL;
+    new->num_ff_side_data = 0;
+    new->ff_side_data = NULL;
     return new;
 }
 
@@ -470,9 +478,8 @@ static void mp_image_copy_cb(struct mp_image *dst, struct mp_image *src,
         memcpy_pic_cb(dst->planes[n], src->planes[n], line_bytes, plane_h,
                       dst->stride[n], src->stride[n], cpy);
     }
-    // Watch out for AV_PIX_FMT_FLAG_PSEUDOPAL retardation
-    if ((dst->fmt.flags & MP_IMGFLAG_PAL) && dst->planes[1] && src->planes[1])
-        memcpy(dst->planes[1], src->planes[1], MP_PALETTE_SIZE);
+    if (dst->fmt.flags & MP_IMGFLAG_PAL)
+        memcpy(dst->planes[1], src->planes[1], AVPALETTE_SIZE);
 }
 
 void mp_image_copy(struct mp_image *dst, struct mp_image *src)
@@ -480,35 +487,10 @@ void mp_image_copy(struct mp_image *dst, struct mp_image *src)
     mp_image_copy_cb(dst, src, memcpy);
 }
 
-void mp_image_copy_gpu(struct mp_image *dst, struct mp_image *src)
+static enum mp_csp mp_image_params_get_forced_csp(struct mp_image_params *params)
 {
-#if HAVE_SSE4_INTRINSICS
-    if (av_get_cpu_flags() & AV_CPU_FLAG_SSE4) {
-        mp_image_copy_cb(dst, src, gpu_memcpy);
-        return;
-    }
-#endif
-    mp_image_copy(dst, src);
-}
-
-// Helper, only for outputting some log info.
-void mp_check_gpu_memcpy(struct mp_log *log, bool *once)
-{
-    if (once) {
-        if (*once)
-            return;
-        *once = true;
-    }
-
-    bool have_sse = false;
-#if HAVE_SSE4_INTRINSICS
-    have_sse = av_get_cpu_flags() & AV_CPU_FLAG_SSE4;
-#endif
-    if (have_sse) {
-        mp_verbose(log, "Using SSE4 memcpy\n");
-    } else {
-        mp_warn(log, "Using fallback memcpy (slow)\n");
-    }
+    int imgfmt = params->hw_subfmt ? params->hw_subfmt : params->imgfmt;
+    return mp_imgfmt_get_forced_csp(imgfmt);
 }
 
 void mp_image_copy_attributes(struct mp_image *dst, struct mp_image *src)
@@ -519,27 +501,21 @@ void mp_image_copy_attributes(struct mp_image *dst, struct mp_image *src)
     dst->dts = src->dts;
     dst->pkt_duration = src->pkt_duration;
     dst->params.rotate = src->params.rotate;
-    dst->params.stereo_in = src->params.stereo_in;
-    dst->params.stereo_out = src->params.stereo_out;
-    if (dst->w == src->w && dst->h == src->h) {
-        dst->params.p_w = src->params.p_w;
-        dst->params.p_h = src->params.p_h;
-    }
-    dst->params.color.primaries = src->params.color.primaries;
-    dst->params.color.gamma = src->params.color.gamma;
-    dst->params.color.sig_peak = src->params.color.sig_peak;
-    dst->params.color.light = src->params.color.light;
-    if ((dst->fmt.flags & MP_IMGFLAG_YUV) == (src->fmt.flags & MP_IMGFLAG_YUV)) {
-        dst->params.color.space = src->params.color.space;
-        dst->params.color.levels = src->params.color.levels;
-        dst->params.chroma_location = src->params.chroma_location;
-    }
+    dst->params.stereo3d = src->params.stereo3d;
+    dst->params.p_w = src->params.p_w;
+    dst->params.p_h = src->params.p_h;
+    dst->params.color = src->params.color;
+    dst->params.chroma_location = src->params.chroma_location;
     dst->params.spherical = src->params.spherical;
-    mp_image_params_guess_csp(&dst->params); // ensure colorspace consistency
+    dst->nominal_fps = src->nominal_fps;
+    // ensure colorspace consistency
+    if (mp_image_params_get_forced_csp(&dst->params) !=
+        mp_image_params_get_forced_csp(&src->params))
+        dst->params.color = (struct mp_colorspace){0};
     if ((dst->fmt.flags & MP_IMGFLAG_PAL) && (src->fmt.flags & MP_IMGFLAG_PAL)) {
         if (dst->planes[1] && src->planes[1]) {
             if (mp_image_make_writeable(dst))
-                memcpy(dst->planes[1], src->planes[1], MP_PALETTE_SIZE);
+                memcpy(dst->planes[1], src->planes[1], AVPALETTE_SIZE);
         }
     }
     av_buffer_unref(&dst->icc_profile);
@@ -651,21 +627,23 @@ char *mp_image_params_to_str_buf(char *b, size_t bs,
         mp_snprintf_cat(b, bs, " %s", mp_imgfmt_to_name(p->imgfmt));
         if (p->hw_subfmt)
             mp_snprintf_cat(b, bs, "[%s]", mp_imgfmt_to_name(p->hw_subfmt));
-        mp_snprintf_cat(b, bs, " %s/%s/%s/%s",
+        if (p->hw_flags)
+            mp_snprintf_cat(b, bs, "[0x%x]", p->hw_flags);
+        mp_snprintf_cat(b, bs, " %s/%s/%s/%s/%s",
                         m_opt_choice_str(mp_csp_names, p->color.space),
                         m_opt_choice_str(mp_csp_prim_names, p->color.primaries),
                         m_opt_choice_str(mp_csp_trc_names, p->color.gamma),
-                        m_opt_choice_str(mp_csp_levels_names, p->color.levels));
+                        m_opt_choice_str(mp_csp_levels_names, p->color.levels),
+                        m_opt_choice_str(mp_csp_light_names, p->color.light));
         if (p->color.sig_peak)
             mp_snprintf_cat(b, bs, " SP=%f", p->color.sig_peak);
         mp_snprintf_cat(b, bs, " CL=%s",
                         m_opt_choice_str(mp_chroma_names, p->chroma_location));
         if (p->rotate)
             mp_snprintf_cat(b, bs, " rot=%d", p->rotate);
-        if (p->stereo_in > 0 || p->stereo_out > 0) {
-            mp_snprintf_cat(b, bs, " stereo=%s/%s",
-                            MP_STEREO3D_NAME_DEF(p->stereo_in, "?"),
-                            MP_STEREO3D_NAME_DEF(p->stereo_out, "?"));
+        if (p->stereo3d > 0) {
+            mp_snprintf_cat(b, bs, " stereo=%s",
+                            MP_STEREO3D_NAME_DEF(p->stereo3d, "?"));
         }
         if (p->spherical.type != MP_SPHERICAL_NONE) {
             const float *a = p->spherical.ref_angles;
@@ -722,13 +700,13 @@ bool mp_image_params_equal(const struct mp_image_params *p1,
 {
     return p1->imgfmt == p2->imgfmt &&
            p1->hw_subfmt == p2->hw_subfmt &&
+           p1->hw_flags == p2->hw_flags &&
            p1->w == p2->w && p1->h == p2->h &&
            p1->p_w == p2->p_w && p1->p_h == p2->p_h &&
            mp_colorspace_equal(p1->color, p2->color) &&
            p1->chroma_location == p2->chroma_location &&
            p1->rotate == p2->rotate &&
-           p1->stereo_in == p2->stereo_in &&
-           p1->stereo_out == p2->stereo_out &&
+           p1->stereo3d == p2->stereo3d &&
            mp_spherical_equal(&p1->spherical, &p2->spherical);
 }
 
@@ -742,7 +720,7 @@ void mp_image_set_attributes(struct mp_image *image,
     nparams.w = image->w;
     nparams.h = image->h;
     if (nparams.imgfmt != params->imgfmt)
-        mp_image_params_guess_csp(&nparams);
+        nparams.color = (struct mp_colorspace){0};
     mp_image_set_params(image, &nparams);
 }
 
@@ -751,12 +729,7 @@ void mp_image_set_attributes(struct mp_image *image,
 // the colorspace as implied by the pixel format.
 void mp_image_params_guess_csp(struct mp_image_params *params)
 {
-    int imgfmt = params->hw_subfmt ? params->hw_subfmt : params->imgfmt;
-    struct mp_imgfmt_desc fmt = mp_imgfmt_get_desc(imgfmt);
-    if (!fmt.id)
-        return;
-
-    enum mp_csp forced_csp = mp_imgfmt_get_forced_csp(imgfmt);
+    enum mp_csp forced_csp = mp_image_params_get_forced_csp(params);
     if (forced_csp == MP_CSP_AUTO) { // YUV/other
         if (params->color.space != MP_CSP_BT_601 &&
             params->color.space != MP_CSP_BT_709 &&
@@ -839,6 +812,13 @@ void mp_image_params_guess_csp(struct mp_image_params *params)
         }
     }
 
+    if (params->chroma_location == MP_CHROMA_AUTO) {
+        if (params->color.levels == MP_CSP_LEVELS_TV)
+            params->chroma_location = MP_CHROMA_LEFT;
+        if (params->color.levels == MP_CSP_LEVELS_PC)
+            params->chroma_location = MP_CHROMA_CENTER;
+    }
+
     if (params->color.light == MP_CSP_LIGHT_AUTO) {
         // HLG is always scene-referred (using its own OOTF), everything else
         // we assume is display-refered by default.
@@ -850,11 +830,17 @@ void mp_image_params_guess_csp(struct mp_image_params *params)
     }
 }
 
-// Copy properties and data of the AVFrame into the mp_image, without taking
-// care of memory management issues.
-static void mp_image_copy_fields_from_av_frame(struct mp_image *dst,
-                                               struct AVFrame *src)
+// Create a new mp_image reference to av_frame.
+struct mp_image *mp_image_from_av_frame(struct AVFrame *src)
 {
+    struct mp_image *dst = &(struct mp_image){0};
+    AVFrameSideData *sd;
+
+    for (int p = 0; p < MP_MAX_PLANES; p++)
+        dst->bufs[p] = src->buf[p];
+
+    dst->hwctx = src->hw_frames_ctx;
+
     mp_image_setfmt(dst, pixfmt2imgfmt(src->format));
     mp_image_set_size(dst, src->width, src->height);
 
@@ -876,11 +862,6 @@ static void mp_image_copy_fields_from_av_frame(struct mp_image *dst,
     if (src->repeat_pict == 1)
         dst->fields |= MP_IMGFIELD_REPEAT_FIRST;
 
-    if (src->hw_frames_ctx) {
-        AVHWFramesContext *fctx = (void *)src->hw_frames_ctx->data;
-        dst->params.hw_subfmt = pixfmt2imgfmt(fctx->sw_format);
-    }
-
     dst->params.color = (struct mp_colorspace){
         .space = avcol_spc_to_mp_csp(src->colorspace),
         .levels = avcol_range_to_mp_csp_levels(src->color_range),
@@ -890,21 +871,86 @@ static void mp_image_copy_fields_from_av_frame(struct mp_image *dst,
 
     dst->params.chroma_location = avchroma_location_to_mp(src->chroma_location);
 
-#if HAVE_OPAQUE_REF
     if (src->opaque_ref) {
         struct mp_image_params *p = (void *)src->opaque_ref->data;
         dst->params.rotate = p->rotate;
-        dst->params.stereo_in = p->stereo_in;
-        dst->params.stereo_out = p->stereo_out;
+        dst->params.stereo3d = p->stereo3d;
+        dst->params.spherical = p->spherical;
+        // Might be incorrect if colorspace changes.
+        dst->params.color.light = p->color.light;
+    }
+
+#if LIBAVUTIL_VERSION_MICRO >= 100
+    sd = av_frame_get_side_data(src, AV_FRAME_DATA_ICC_PROFILE);
+    if (sd)
+        dst->icc_profile = sd->buf;
+
+    // Get the content light metadata if available
+    sd = av_frame_get_side_data(src, AV_FRAME_DATA_CONTENT_LIGHT_LEVEL);
+    if (sd) {
+        AVContentLightMetadata *clm = (AVContentLightMetadata *)sd->data;
+        dst->params.color.sig_peak = clm->MaxCLL / MP_REF_WHITE;
+    }
+
+    // Otherwise, try getting the mastering metadata if available
+    sd = av_frame_get_side_data(src, AV_FRAME_DATA_MASTERING_DISPLAY_METADATA);
+    if (!dst->params.color.sig_peak && sd) {
+        AVMasteringDisplayMetadata *mdm = (AVMasteringDisplayMetadata *)sd->data;
+        if (mdm->has_luminance)
+            dst->params.color.sig_peak = av_q2d(mdm->max_luminance) / MP_REF_WHITE;
+    }
+
+    sd = av_frame_get_side_data(src, AV_FRAME_DATA_A53_CC);
+    if (sd)
+        dst->a53_cc = sd->buf;
+
+    for (int n = 0; n < src->nb_side_data; n++) {
+        sd = src->side_data[n];
+        struct mp_ff_side_data mpsd = {
+            .type = sd->type,
+            .buf = sd->buf,
+        };
+        MP_TARRAY_APPEND(NULL, dst->ff_side_data, dst->num_ff_side_data, mpsd);
     }
 #endif
+
+    if (dst->hwctx) {
+        AVHWFramesContext *fctx = (void *)dst->hwctx->data;
+        dst->params.hw_subfmt = pixfmt2imgfmt(fctx->sw_format);
+        const struct hwcontext_fns *fns =
+            hwdec_get_hwcontext_fns(fctx->device_ctx->type);
+        if (fns && fns->complete_image_params)
+            fns->complete_image_params(dst);
+    }
+
+    struct mp_image *res = mp_image_new_ref(dst);
+
+    // Allocated, but non-refcounted data.
+    talloc_free(dst->ff_side_data);
+
+    return res;
 }
 
-// Copy properties and data of the mp_image into the AVFrame, without taking
-// care of memory management issues.
-static void mp_image_copy_fields_to_av_frame(struct AVFrame *dst,
-                                             struct mp_image *src)
+
+// Convert the mp_image reference to a AVFrame reference.
+struct AVFrame *mp_image_to_av_frame(struct mp_image *src)
 {
+    struct mp_image *new_ref = mp_image_new_ref(src);
+    AVFrame *dst = av_frame_alloc();
+    if (!dst || !new_ref) {
+        talloc_free(new_ref);
+        av_frame_free(&dst);
+        return NULL;
+    }
+
+    for (int p = 0; p < MP_MAX_PLANES; p++) {
+        dst->buf[p] = new_ref->bufs[p];
+        new_ref->bufs[p] = NULL;
+    }
+
+    dst->hw_frames_ctx = new_ref->hwctx;
+    new_ref->hwctx = NULL;
+
     dst->format = imgfmt2pixfmt(src->imgfmt);
     dst->width = src->w;
     dst->height = src->h;
@@ -934,45 +980,48 @@ static void mp_image_copy_fields_to_av_frame(struct AVFrame *dst,
 
     dst->chroma_location = mp_chroma_location_to_av(src->params.chroma_location);
 
-#if HAVE_OPAQUE_REF
-    av_buffer_unref(&dst->opaque_ref);
     dst->opaque_ref = av_buffer_alloc(sizeof(struct mp_image_params));
     if (!dst->opaque_ref)
         abort();
     *(struct mp_image_params *)dst->opaque_ref->data = src->params;
-#endif
-}
 
-// Create a new mp_image reference to av_frame.
-struct mp_image *mp_image_from_av_frame(struct AVFrame *av_frame)
-{
-    struct mp_image t = {0};
-    mp_image_copy_fields_from_av_frame(&t, av_frame);
-    for (int p = 0; p < MP_MAX_PLANES; p++)
-        t.bufs[p] = av_frame->buf[p];
-    t.hwctx = av_frame->hw_frames_ctx;
-    return mp_image_new_ref(&t);
-}
-
-// Convert the mp_image reference to a AVFrame reference.
-struct AVFrame *mp_image_to_av_frame(struct mp_image *img)
-{
-    struct mp_image *new_ref = mp_image_new_ref(img);
-    AVFrame *frame = av_frame_alloc();
-    if (!frame || !new_ref) {
-        talloc_free(new_ref);
-        av_frame_free(&frame);
-        return NULL;
+#if LIBAVUTIL_VERSION_MICRO >= 100
+    if (src->icc_profile) {
+        AVFrameSideData *sd =
+            av_frame_new_side_data_from_buf(dst, AV_FRAME_DATA_ICC_PROFILE,
+                                            new_ref->icc_profile);
+        if (!sd)
+            abort();
+        new_ref->icc_profile = NULL;
     }
-    mp_image_copy_fields_to_av_frame(frame, new_ref);
-    for (int p = 0; p < MP_MAX_PLANES; p++)
-        frame->buf[p] = new_ref->bufs[p];
-    frame->hw_frames_ctx = new_ref->hwctx;
-    *new_ref = (struct mp_image){0};
+
+    if (src->params.color.sig_peak) {
+        AVContentLightMetadata *clm =
+            av_content_light_metadata_create_side_data(dst);
+        if (!clm)
+            abort();
+        clm->MaxCLL = src->params.color.sig_peak * MP_REF_WHITE;
+    }
+
+    // Add back side data, but only for types which are not specially handled
+    // above. Keep in mind that the types above will be out of sync anyway.
+    for (int n = 0; n < new_ref->num_ff_side_data; n++) {
+        struct mp_ff_side_data *mpsd = &new_ref->ff_side_data[n];
+        if (!av_frame_get_side_data(dst, mpsd->type)) {
+            AVFrameSideData *sd = av_frame_new_side_data_from_buf(dst, mpsd->type,
+                                                                  mpsd->buf);
+            if (!sd)
+                abort();
+            mpsd->buf = NULL;
+        }
+    }
+#endif
+
     talloc_free(new_ref);
-    if (frame->format == AV_PIX_FMT_NONE)
-        av_frame_free(&frame);
-    return frame;
+
+    if (dst->format == AV_PIX_FMT_NONE)
+        av_frame_free(&dst);
+    return dst;
 }
 
 // Same as mp_image_to_av_frame(), but unref img. (It does so even on failure.)
